<?php

namespace Drupal\search_api_solr_boosted_keyword\EventSubscriber;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\search_api\Entity\Index;
use Drupal\search_api_solr\Event\PostCreateIndexDocumentsEvent;
use Drupal\search_api_solr\Event\PreQueryEvent;
use Drupal\search_api_solr\Event\SearchApiSolrEvents;
use Drupal\search_api_solr\SolrBackendInterface;
use Drupal\search_api_solr\Utility\Utility;
use Solarium\QueryType\Select\Query\Query;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Event subscriber for Search API Solr events.
 */
class SearchApiSolrBoostedKeywordEventSubscriber implements EventSubscriberInterface {

  /**
   * The language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * Constructs a new SearchApiSolrBoostedKeywordEventSubscriber object.
   *
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   The language manager.
   */
  public function __construct(LanguageManagerInterface $language_manager) {
    $this->languageManager = $language_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return [
      SearchApiSolrEvents::POST_CREATE_INDEX_DOCUMENTS => 'alterSolrDocuments',
      SearchApiSolrEvents::PRE_QUERY => 'alterSolrQuery',
    ];
  }

  /**
   * Alters Solr documents before they are sent to Solr for indexing.
   *
   * @param \Drupal\search_api_solr\Event\PostCreateIndexDocumentsEvent $event
   *   The event.
   *
   * @throws \Drupal\search_api\SearchApiException
   */
  public function alterSolrDocuments(PostCreateIndexDocumentsEvent $event) {
    $documents = $event->getSolariumDocuments();
    $document = reset($documents);
    $index_id = $document->getFields()['index_id'];
    $index = Index::load($index_id);
    $items = $event->getSearchApiItems();

    // Collect all indexed boosted keyword fields.
    $keywordFields = [];
    $fields = $index->getFields();
    $solrBackend = $index->getServerInstance()->getBackend();
    $solrFields = [];
    if ($solrBackend instanceof SolrBackendInterface) {
      $languages = $this->languageManager->getLanguages();
      $languageIds = array_map(
        function (LanguageInterface $value) {
          return $value->getId();
        }, $languages);
      $solrFields = $solrBackend->getSolrFieldNamesKeyedByLanguage($languageIds, $index);
    }
    foreach ($fields as $fieldName => $field) {
      if ($field->getOriginalType() == 'field_item:search_api_solr_boosted_keyword') {
        if (!empty($solrFields[$fieldName])) {
          foreach ($solrFields[$fieldName] as $langcode => $solrField) {
            $keywordFields[$langcode][$solrField] = $field->getPropertyPath();
          }
        }
      }
    }

    // If no keyword fields were found, do nothing.
    if (empty($keywordFields)) {
      return;
    }

    // Go over all documents to alter boosted keyword indexing.
    /** @var \Solarium\Core\Query\DocumentInterface $document */
    foreach ($documents as $document) {
      // Get all document fields.
      $docFields = $document->getFields();

      // Find the corresponding item to get to the content entity.
      /** @var \Drupal\search_api\Item\Item $item */
      $item = (!empty($items[$docFields['ss_search_api_id']])) ? $items[$docFields['ss_search_api_id']] : NULL;
      if (!empty($item)) {
        /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
        $entity = $item->getOriginalObject()->getValue();
        $langcode = $item->getLanguage();
        if (!empty($entity) && $entity instanceof ContentEntityInterface && !empty($keywordFields[$langcode])) {
          // Alter the indexed value for all keyword fields.
          foreach ($keywordFields[$langcode] as $solrFieldName => $fieldName) {
            if ($entity->hasField($fieldName)) {
              $fieldValues = $entity->get($fieldName)->getValue();
              if (!empty($fieldValues)) {
                $indexValue = $this->getKeywordIndexValue($docFields[$solrFieldName], $fieldValues);
                $document->setField($solrFieldName, $indexValue);
              }
            }
          }
        }
      }
    }

    $event->setSolariumDocuments($documents);
  }

  /**
   * Alters Solr queries before they are sent to Solr.
   *
   * @param \Drupal\search_api_solr\Event\PreQueryEvent $event
   *   The event.
   *
   * @throws \Drupal\search_api\SearchApiException
   */
  public function alterSolrQuery(PreQueryEvent $event) {
    $solarium_query = $event->getSolariumQuery();
    $query = $event->getSearchApiQuery();

    // Get the indexed boosted keyword fields and their boost levels.
    $index = $query->getIndex();
    $indexFields = $index->getFields();
    $solrBackend = $index->getServerInstance()->getBackend();
    $solrFields = [];
    if ($solrBackend instanceof SolrBackendInterface) {
      $languages = $this->languageManager->getLanguages();
      $languageIds = array_map(
        function (LanguageInterface $value) {
          return $value->getId();
        }, $languages);
      $solrFields = $solrBackend->getSolrFieldNamesKeyedByLanguage($languageIds, $index);
    }
    $keywordFields = [];
    foreach ($indexFields as $fieldName => $field) {
      if ($field->getOriginalType() == 'field_item:search_api_solr_boosted_keyword') {
        if (!empty($solrFields[$fieldName])) {
          if (is_array($solrFields[$fieldName])) {
            $languages = $query->getLanguages();
            foreach ($languages as $langcode) {
              if (!empty($solrFields[$fieldName][$langcode])) {
                $languageSpecificName = Utility::getLanguageSpecificSolrDynamicFieldNameForSolrDynamicFieldName(
                  $solrFields[$fieldName][$langcode],
                  $langcode
                );
                $keywordFields[$languageSpecificName] = $field->getBoost();
              }
            }
          }
          else {
            $keywordFields[$solrFields[$fieldName]] = $field->getBoost();
          }
        }
      }
    }

    // Set up the boost functions.
    $keys = $query->getKeys();
    if (!empty($keys)) {
      $keys = (!is_array($keys)) ? [$keys] : $keys;
      foreach ($keys as $keyIndex => $key) {
        if (substr($keyIndex, 0, 1) !== '#') {
          // Divide phrases in to words because termfreq does not work with
          // phrases.
          $keyParts = explode(' ', trim($key));
          foreach ($keyParts as &$keyPart) {
            $keyPart = $solarium_query->getHelper()->escapeTerm($keyPart);
            foreach ($keywordFields as $fieldName => $boost) {
              $boostFunctions[] = 'termfreq("' . $fieldName . '","' . $keyPart . '")^' . $boost;
            }
          }
          // If we had a phrase, add a more important boost function for the
          // phrase without spaces (see indexing).
          if (count($keyParts) > 1) {
            $wordBoost = intval($boost) * count($keyParts);
            $boostFunctions[] = 'termfreq("' . $fieldName . '","' . implode('', $keyParts) . '")^' . $wordBoost;
          }
        }
      }
    }

    // Apply the boost functions if possible.
    if ($solarium_query instanceof Query) {
      $edIsMax = $solarium_query->getEDisMax();
      if (!empty($boostFunctions)) {
        $edIsMax->setBoostFunctions(implode(' ', $boostFunctions));

        // Avoid the conversion into a lucene parser expression, keep edismax.
        $solarium_query->addParam('defType', 'edismax');

        // Exclude all boosted fields from the search query.
        $queryFields = $edIsMax->getQueryFields();
        foreach ($keywordFields as $keywordField => $boost) {
          $queryFields = preg_replace('/' . $keywordField . '\^\d+/', '', $queryFields);
        }
        $edIsMax->setQueryFields($queryFields);
      }
    }
  }

  /**
   * Returns the value to index for a boosted keyword field.
   *
   * Based on the currently indexed value and the field values.
   * We make sure to strip spaces for phrase support as termfreq boost functions
   * don't work with phrases.
   *
   * @param mixed $originalValue
   *   The original value to be indexed.
   * @param array $fieldValues
   *   The field values containing keywords and boost levels.
   *
   * @return array|string
   *   The processed value to be indexed.
   */
  protected function getKeywordIndexValue($originalValue, $fieldValues) {
    if (is_array($originalValue)) {
      $indexValue = [];
      foreach ($fieldValues as $fieldValue) {
        $boostLevel = intval($fieldValue['boost']);
        for ($i = 0; $i < $boostLevel; $i++) {
          $indexValue[] = str_replace(' ', '', $fieldValue['value']);
        }
      }
    }
    else {
      $indexValue = [];
      $boostLevel = intval($fieldValues[0]['boost']);
      for ($i = 0; $i < $boostLevel; $i++) {
        $indexValue[] = str_replace(' ', '', $fieldValues[0]['value']);
      }
      $indexValue = implode(' ', $indexValue);
    }

    return $indexValue;
  }

}
